/***************************************************************************
   network_icmp.c  -  ICMP network plugin
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: network_icmp.c,v 1.33 2003/12/06 00:14:04 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file network_icmp.c
  * File contains ICMP plugin functions
  * This file is a good example of the plugin API.
  * Use it as an example for your own plugin development.
  */

#include <assert.h>
#include <pthread.h>
#include "network_icmp.h"
#include "error.h"
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif

static const char rcsid[] =
        "\100$ autosec/sonar: $Id: network_icmp.c,v 1.33 2003/12/06 00:14:04 red0x Exp $";

static unsigned char done = 0;

/// my last error
static uint16_t my_errno;
/// my socket
static int my_socket,
/// my sequence number
	seq_no, seq_no2,
/// lowest sequence number
  low_seq,
/// my ICMP type
  my_type,
/// my ICMP code
  my_code;

/// ready for use
static bool ready;
/// payload file name
static char *payload_file;
/// Timestamp Time zone
static struct timezone tz;
/// Timestamp timeval
static struct timeval tp;
static unsigned int num_hosts;

/// Plugin's plugin_t structure, built by plugin_init
static plugin_t pthis;
/// Sonar's sonar_t structure, built by sonar, passed to this plugin
static sonar_t *mthis;

static int icmp_end(void);
static int icmp_args(char opt, char *optarg);
static int icmp_leadin(void *c_data, int i_data);
static int icmp_close(void);
static int icmp_sweep(target_t *targets);
static void icmp_usage(void);
static void icmp_status(void);
static int icmp_ping(target_t *target);
static void *recv_thread(void *arg);
static unsigned short checksum(unsigned short *buffer, int size);
static int icmp_last_error(void);
static void inline error(char *x, int y);
static int resptype(int type);
static const char *respstr(int type);
static const char *unreach_codes(int code);
static RETSIGTYPE sig_handler(int i);

/** @fn plugin_init(void *in_data)
  * Plugin Init function, part of the Plugin API
  * @param in_data Pointer to a sonar_t structure that will be 
  *  passed from sonar when your plugin is loaded.
  * @return A pointer to a filled plugin_t structure.
  */
void *plugin_init(void * in_data)
{
        assert(in_data != NULL);
        //useless reference, to avoid a warning
        if(rcsid)
        {
        }
        // this is assumed
        mthis = (sonar_t *) in_data;
        if(mthis->api < 5)
                err_msg("Plugin API must be 5 or higher, update sonar");

        srand(getpid() ^ (uid_t) time(NULL));
        my_errno = PLUGIN_LOADED;
        my_socket = -1;

        seq_no = (int) (256.0 * rand() / (RAND_MAX + 0.0));
        payload_file = NULL;
	pthis.version = 5; /* set API version 5 */

        //setup my plugin stuff
        pthis.type = NETWORK_PLUGIN;
        pthis.err = (int16_t *) &my_errno;
        pthis.ready = false;

        pthis.plugin_char = 'I';
        pthis.shortname = "icmp scan";
        pthis.longname = "ICMP Scanning Plugin";

        pthis.plugin_init = &plugin_init;
        pthis.plugin_destroy = &icmp_end;
        pthis.plugin_open = &icmp_leadin;
        pthis.plugin_close = &icmp_close;
        pthis.plugin_next = NULL;
        pthis.plugin_sweep = &icmp_sweep;
        pthis.plugin_args = &icmp_args;
        pthis.plugin_sweep = &icmp_sweep;
        pthis.plugin_output = NULL;
	pthis.plugin_results = NULL;
        pthis.plugin_usage = &icmp_usage;
        pthis.plugin_status = &icmp_status;
        pthis.plugin_last_error = &icmp_last_error;
        ready = false;

        num_hosts = 0;
        my_type = ICMP_ECHO;
        my_code = 0;

        return &pthis;
}

/** @fn icmp_end(void)
  * Plugin Unload function
  * @return PLUGIN_UNLOADED
  */
static int icmp_end(void)
{
        my_errno = PLUGIN_UNLOADED;
        if(my_socket != -1 && my_socket != 0)
        {
                close(my_socket);
                my_socket = -1;
        }

        seq_no = 0;

        if(payload_file != NULL)
        {
                free(payload_file);
                payload_file = NULL;
        }

        pthis.type = NETWORK_PLUGIN;
        pthis.err = (int16_t *) &my_errno;
        pthis.ready = false;
        pthis.plugin_init = NULL;
        pthis.plugin_destroy = NULL;
        pthis.plugin_open = NULL;
        pthis.plugin_close = NULL;
        pthis.plugin_next = NULL;
        pthis.plugin_sweep = NULL;
        pthis.plugin_args = NULL;
        pthis.plugin_sweep = NULL;
        pthis.plugin_output = NULL;
	pthis.plugin_results = NULL;
        pthis.plugin_usage = NULL;
        pthis.plugin_status = NULL;
        pthis.plugin_last_error = NULL;
        ready = false;

        num_hosts = 0;
        mthis = NULL;
        return PLUGIN_UNLOADED;
}


/** @fn icmp_args(char opt, char *optarg)
  * Plugin Argument handler
  * @param opt Option char as passed from getopt
  * @param optarg Option argument as passed from getopt
  * @return PLUGIN_OK if that argument was handled, PLUGIN_ERROR otherwise
  */
static int icmp_args(char opt, char *optarg)
{
        int mlen;
        
	assert(mthis != NULL);
	assert(mthis->memput != NULL);
        

        switch (opt)
        {
                case 'p':
                        mthis->memput(&payload_file, optarg, strlen(optarg));
                        break;
                case 'i':
                        mlen = strlen(optarg);
                        my_type = -1;
                        if(strncmp(optarg, "echo", (mlen < 4) ? mlen : 4) ==
                           0)

                                my_type = ICMP_ECHO;
                        else if(strncmp(optarg, "du", (mlen < 2) ? mlen : 2)
                                == 0)
                                my_type = ICMP_UNREACH;
                        else if(strncmp(optarg, "mask", (mlen < 4) ? mlen : 4)
                                == 0)
                                my_type = ICMP_MASKREQ;
                        else if(strncmp(optarg, "red", (mlen < 3) ? mlen : 3)
                                == 0)
                                my_type = ICMP_REDIRECT;
                        else if(strncmp(optarg, "ts", (mlen < 2) ? mlen : 2)
                                == 0)
                                my_type = ICMP_TSTAMP;
                        else if(strncmp(optarg, "info", (mlen < 4) ? mlen : 4)
                                == 0)
                                my_type = ICMP_IREQ;
                        else if(strncmp
                                (optarg, "param", (mlen < 5) ? mlen : 5) == 0)
                                my_type = ICMP_PARAMPROB;
                        else if(strncmp(optarg, "help", (mlen < 4) ? mlen : 4)
                                == 0)
                        {
                                err_quit("Currently supported types are:\n"
                                         "\t echo - echo request\n"
                                         "\t du - dest unreachable\n"
                                         "\t mask - mask request\n"
                                         "\t red - redirect\n"
                                         "\t ts - timestamp request\n"
                                         "\t info - info request\n"
                                         "\t param - param problem\n");
                        }
                        else
                        {
                                err_quit("type not supported\n"
                                         "Currently supported types are:\n"
                                         "\t echo - echo request\n"
                                         "\t du - dest unreachable\n"
                                         "\t mask - mask request\n"
                                         "\t red - redirect\n"
                                         "\t ts - timestamp request\n"
                                         "\t info - info request\n"
                                         "\t param - param problem\n");
                        }
                        if(!ICMP_INFOTYPE(my_type)
                           && my_type != ICMP_PARAMPROB
                           && my_type != ICMP_UNREACH
                           && my_type != ICMP_REDIRECT)
                        {
                                err_quit("type not supported");
                        }
                        break;
                case 'c':
                        mlen = strlen(optarg);
                        my_code = atoi(optarg);
                        break;

                default:
                        //we dont handle anything else
                        return PLUGIN_ERROR;
        }
        // we handled it
        return PLUGIN_OK;
}

/** @fn icmp_leadin(void *c_data, int i_data)
  * Plugin Start function
  * @param c_data Not used yet
  * @param i_data Not used yet
  * @return PLUGIN_OK
  */
static int icmp_leadin(void *c_data, int i_data)
{
        // load socket stuff here
        // i_data will be the TTL to set it to
        uid_t uid;
        struct linger l;
        
	assert(mthis != NULL);
	assert(mthis->usage != NULL);
	assert(mthis->getprivs != NULL);
	assert(mthis->dropprivs != NULL);
        
        if(mthis->pCon.iPostDelay == -1)
                mthis->pCon.iPostDelay = 0;     //default Post Delay of (NONE)
        // setuid root section of code
        {
                uid = mthis->getprivs();
                // root section of code
                //get me a RAW ICMP socket
                my_socket = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
                if(my_socket == -1)
                {
                        err_ret("could not get raw socket\n");
                        exit(-1);
                }
                if(mthis->pCon.iTTL > 0)
                        if(setsockopt
                           (my_socket, SOL_IP, IP_TTL, &(mthis->pCon.iTTL),
                            sizeof(mthis->pCon.iTTL)) < 0)
                        {
                                err_ret("could not set TTL");
                                //perror("setsockopt");
                        }
                l.l_onoff = 1;
                l.l_linger = mthis->pCon.iTimeout;
                if(setsockopt(my_socket, SOL_SOCKET, SO_LINGER,
                              (struct linger *) &l,
                              sizeof(struct linger)) < 0)
                {
                        err_ret("could not set timeout");
                        //perror("setsockopt");
                }
                if(mthis->pCon.bSockDebug)
                {
                        l.l_onoff = 1;
                        if(setsockopt(my_socket, SOL_SOCKET, SO_DEBUG,
                                      (int *) &(l.l_onoff), sizeof(int)) < 0)
                        {
                                err_ret("could not set debug mode");
                                //perror("setsockopt");
                        }
                }
                if(mthis->pCon.bBroadcast)
                {
                        l.l_onoff = 1;
                        if(setsockopt(my_socket, SOL_SOCKET, SO_BROADCAST,
                                      (int *) &l.l_onoff, sizeof(int)) < 0)
                        {
                                err_ret("could not set broadcast mode");
                                //perror("setsockopt");
                        }
                }
                mthis->dropprivs(uid);
        }
        //back to normalcy
        ready = true;
	pthis.ready = true;
        seq_no = 1 + (int) (70.0 * rand() / (RAND_MAX + 1.0));
	num_hosts = mthis->num_targets;
        return PLUGIN_OK;
}

/** @fn icmp_close(void)
  * Plugin destroy function
  * Call this before icmp_end
  * @return PLUGIN_OK
  */
static int icmp_close(void)
{
        close(my_socket);
        my_socket = -1;
        ready = false;
        payload_file = NULL;
        pthis.ready = false;
        return PLUGIN_OK;
}

/** @fn icmp_sweep(target_t *targets)
  * Plugin Sweep Function
  * Runs a network sweep (this is for network plugins only).
  * Recompiled to make one pass only!!!
  * @param targets Pointer to linked list of targets
  * @return PLUGIN_ERROR on error, PLUGIN_OK on ok
  */
static int icmp_sweep(target_t *targets)
{
        unsigned int counter = 0;
        pthread_t tid;
        pthread_attr_t attr;
        struct recv_t recd;
        int timeout;

	assert(targets != NULL);

        // just make sure we, and sonar, are ready for processing
        if(!ready || !pthis.ready)
        {
                err_msg("plugin not ready\n");
                return PLUGIN_ERROR;
        }

        timeout = mthis->pCon.iTimeout;
        if(timeout <= 0)
                timeout = 5000; //default timeout of 5 seconds

        pthread_attr_init(&attr);
        memset(&recd, 0, sizeof(struct recv_t));

        recd.socket = my_socket;
        recd.hosts = num_hosts;
	/* this is a broad cast address, not a range */
	if(mthis->pCon.bBroadcast > 0 && targets->extra_hosts == 0)
		recd.hosts += (unsigned int) mthis->pCon.bBroadcast * 255;
        low_seq = recd.start = clock(); /*seq_no;*/

        /* set the default timeout values */
        recd.max = recd.start + (timeout * recd.hosts * CLOCKS_PER_SEC) / 1000;
	/* add time for a per host delay */
	if(mthis->pCon.iHostDelay > 0)
		recd.max += (mthis->pCon.iHostDelay * recd.hosts * CLOCKS_PER_SEC) / 1000;
	/* add a default magic number to get some extra (late) responses */
	recd.max += CLOCKS_PER_SEC/2000;  /* default is one half extra second */
	
	done = 0;
        if(pthread_create(&tid, &attr, recv_thread, &recd) != 0)
                err_sys("could not create thread");

        while(targets != NULL && counter < num_hosts)
        {
		/* we have an ip-range here */
		if(targets->extra_hosts > 0)
		{
			/* increas the timeout */
			recd.max += (timeout * targets->extra_hosts * CLOCKS_PER_SEC) / 1000;
#ifdef DEBUG
			fprintf(stderr, "%s: increasing recd.hosts by %d\n", targets->entry, targets->extra_hosts);
#endif
			recd.hosts += targets->extra_hosts;
			if(mthis->sCon.iVerbose > 0)
				err_msg("Increasing timeout, netmask detected: %d", recd.max);
		}
                if(icmp_ping(targets) == PLUGIN_ERROR)
                {
                        return PLUGIN_ERROR;
                }
                /*usleep(10);*/     /* wait a while */

                targets = targets->next;
                counter++;
		/* sleep after this host (allows more dest unreachables to come back) */
		if(counter < num_hosts)
			usleep(mthis->pCon.iHostDelay * 1000);
        }
	done = 1;
        /* wait for the thread to exit */
        pthread_join(tid, NULL);

        if(mthis->pCon.iPostDelay)
                usleep(mthis->pCon.iPostDelay * 1000);
        return PLUGIN_OK;
}

/** @fn icmp_status(void)
  * Plugin Status hook function
  * Prints out status information right before the sweep begins
  */
static void icmp_status(void)
{
        if(pthis.ready)
        {
                mthis->output("ICMP sweep: %d host(s)::\n", num_hosts);
                mthis->output("\tType: %s Code: %d (non-verified)\n",
                              respstr(my_type), my_code);
                if(mthis->pCon.iPostDelay == 0)
                        mthis->output("\tno post-delay::\n");
                else
                        mthis->output("\tpost-delay %ds::\n",
                                      mthis->pCon.iPostDelay);
                if(payload_file != NULL)
                        mthis->output("\tpayload: %s::\n", payload_file);
                else
                        mthis->output("\tno payload::\n");
                if(mthis->pCon.iSize != 0)
                        mthis->output("\tsize %d::\n", mthis->pCon.iSize);
                else
                        mthis->output("\trandom packet size::\n");
                if(mthis->pCon.iTTL != 255)
                        mthis->output("\tTTL %d::\n", mthis->pCon.iTTL);
                else
                        mthis->output("\tdefault TTL::\n");
        }
}

/** @fn icmp_usage(void)
  * Plugin Usage hook function
  * Prints out plugin usage information. All options preceded by -p are passed
  * off to plugins, so make your usage reflect that.  However, when your arguments
  * get to your plugin's args function, the p will be stripped.
  */
static void icmp_usage(void)
{
        assert(mthis->output != NULL);
        
        mthis->output("\n ICMP Scan plugin: requires root priveleges\n");
        mthis->output
                ("  Usage: [-pp file] [-pi type] [-pc code] hostname...\n");
        mthis->output
                ("\t  -pp: append contents of file to payload of ICMP packet\n");
        mthis->output("\t  -pi: set ICMP type (-pihelp for a list)\n");
        mthis->output("\t  -pc: set ICMP code\n");
        mthis->output("  Example: sonar -sI example.com\n");
        exit(-1);
}

/** @fn icmp_ping(target_t *target)
  * Plugin Ping function
  * Sends out an ICMP ECHO_REQUEST packet
  * @param target pointer to a host_list_t structure containing current host to ping
  * @return PLUGIN_OK on ok, PLUGIN_ERROR otherwise
  */
static int icmp_ping(target_t *target)
{
        struct icmp *icmp_hdr;
        int total_size = sizeof(struct icmp), dlen = 0;
        int bwrote;
	unsigned int j;
        //struct timezone tz;
        //register struct timeval *tp;
        char *buf, *datapart;
	struct sockaddr_in *sa;
	struct in_addr ia;

	assert(target != NULL);
	assert(target->entry != NULL);
	assert(target->addrinfo != NULL);
	assert(target->addrinfo->ai_addr != NULL);
	assert(my_socket != 0);
	assert(my_socket != -1);

	srand(getpid() ^ (uid_t) time(NULL));
	/* set the packet size to specified */
        if(mthis->pCon.iSize != 0)
                total_size += mthis->pCon.iSize;
	/* or pick a random size 64 to 290*/
        else
                total_size += 64 + (int) (290.0 * rand() / (RAND_MAX + 64.0));

	for(j = 0; j <= target->extra_hosts; j++)
	{
		/* get the right address */
		switch(target->addrinfo->ai_family)
		{
		case PF_INET:
			sa = (struct sockaddr_in *) target->addrinfo->ai_addr;
			ia = sa->sin_addr;
			if(j > 0)
				ia.s_addr = htonl(ntohl(ia.s_addr)+1);
			sa->sin_addr = ia;
			break;
		default:
			err_quit("address family not supported yet\n");
		}
		/* how much of the packet is data? */
		dlen = total_size - sizeof(struct icmp);
		buf = (char *) malloc(total_size);
		if(!buf)
		{
			err_ret("failed to malloc");
			return PLUGIN_ERROR;
		}
		memset(buf, 0, total_size);

		icmp_hdr = (struct icmp *) buf;
		icmp_hdr->icmp_type = my_type;
		icmp_hdr->icmp_code = my_code;
		icmp_hdr->icmp_id = (unsigned short) getpid();
		icmp_hdr->icmp_cksum = 0;
		/* random sequence number, to make fingerprinting this tool harder */
		/* clock works well, that way we can count round trip time as well */
		/* however, the down side is that they always increase... */
		icmp_hdr->icmp_seq = seq_no2 = clock();/*seq_no2 = low_seq + (int) (25.0 * rand() / (RAND_MAX + 1.0));*/

		datapart = buf + sizeof(struct icmp);
		/* no payload */
		if(!payload_file)
		{
			/* do we want a pattern? */
			if(mthis->pCon.sPattern)
			{
				int i, len = strlen(mthis->pCon.sPattern);
				int num = dlen / len;
				for(i = 0; i < num; i++)
				{
					strncpy(datapart, mthis->pCon.sPattern, len);
					datapart += len;
				}
			}
			/* or random garbage */
			else
			{
				FILE *rp;
				rp = fopen("/dev/urandom", "r");
				if(!rp)
				{
	#ifdef DEBUG
					fprintf(stderr, "network_icmp.c: could not open /dev/urandom\n");
	#endif
					rp = fopen("/dev/random", "r");
					if(!rp)
					{
	#ifdef DEBUG
					fprintf(stderr, "network_icmp.c: could not open /dev/random\n");
	#endif
						memset(datapart,
					(int) (256.0 * rand() / (RAND_MAX + 0.0)),
				       dlen);
					}				
				}
				/* if we didn't set it already */
				if(*datapart == 0)
				{
					fread(datapart, dlen, 1, rp);
					fclose(rp);
				}
			}
		}
		/* we've got a payload */
		else
		{
			unsigned int len = 0, cur = 0;
			FILE *payload;
			payload = fopen(payload_file, "r");
			if(!payload)
			{
				*(pthis.err) = PLUGIN_NOFILE;
				err_ret("could not open payload file");
				free(buf);
				return PLUGIN_ERROR;
			}
			fseek(payload, 0, SEEK_END);
			len = (unsigned int) ((ftell(payload) > dlen) ? dlen :
					      ftell(payload));
			fseek(payload, 0, SEEK_SET);
			while(!feof(payload) && cur < len)
			{
				int c = fgetc(payload);
				datapart[cur++] = c;
			}
		}

		gettimeofday(&tp, &(tz));

		((struct icmp *) buf)->icmp_cksum =
			checksum((unsigned short *) buf, total_size);
		bwrote = sendto(my_socket, buf, total_size, 0,
				target->addrinfo->ai_addr, sizeof(struct sockaddr));
		free(buf);
		if(bwrote == -1)
		{
			err_ret("sendto failed");
			//perror("sendto");
			return PLUGIN_ERROR;
		}
		if(bwrote < (int) total_size)
		{
			err_msg("wrote %d bytes\n", bwrote);
		}
		
	}
        return PLUGIN_OK;
}

/** @fn recv_thread(void *arg)
  * Plugin Receive thread
  * Looks for ICMP ECHO_REPLY packets
  * @param arg Pointer to recv_t structure
  * @return This thread returns NULL when its done.
  */
static void *recv_thread(void *arg)     //struct recv_t *rec)
{
        fd_set rfds;
        struct ip *iphdr;
        struct icmp *icmphdr;
        unsigned short iphdrlen;
        unsigned long triptime;
	unsigned long num_res = 0;
        struct timeval tv;
        int bread, ret, len;
        char buf[MAX_PACKET], *adr = NULL;
        struct recv_t *rec = (struct recv_t *) arg;
        int fromlen, my_socket = rec->socket, result;
        struct sockaddr_in from;
	static struct addrinfo *ai = NULL;
	result_t *results,		/* results used to pass back to sonar */
		*first_res;	 	/* first on the list */
	static result_t res;			/* temporary for *this* result */
	int psuedo = false;

	/* signals, baby */
	signal(SIGTERM, sig_handler);
        signal(SIGINT, sig_handler);
        signal(SIGHUP, sig_handler);
	signal(SIGSEGV, sig_handler);
	/* make us a list */
	results = (result_t *) malloc(sizeof(result_t));
	if(!results)
		err_sys("No memory for result list.\n");
	first_res = results; /* store a pointer to the top of the list */
        rec->responses = 0;
	memset(&res, 0, sizeof(result_t));
	memset(results, 0, sizeof(result_t));

        while((done == 0 || clock() <= rec->max) && rec->responses < rec->hosts)
        {
                tv.tv_usec = 0;
                tv.tv_sec = 0;
                
                FD_ZERO(&rfds);
                FD_SET(my_socket, &rfds);

                ret = select(my_socket + 1, &rfds, NULL, NULL, &tv);
                if(ret == -1)
                {
                        *(pthis.err) = PLUGIN_ERROR;
                        err_ret("could not select socket\n");
                        return NULL;    //PLUGIN_ERROR;
                }
                if(ret == 0)
                {
                        continue;
                }
		if(psuedo == true)
                {
                        assert(ai != NULL);
                        psuedo = false;
                        free(ai);
                }
		memset(&from, 0, sizeof(struct sockaddr_in));
		memset(buf, 0, MAX_PACKET);
		
                bread = recvfrom(my_socket, buf, MAX_PACKET, 0,
                                 (struct sockaddr *) &(from),
                                 (socklen_t *) & (fromlen));
                if(bread == -1)
                {
                        err_ret("recvfrom");
                        //perror("recvfrom");
                        continue;
                }

		memset(&res, 0, sizeof(result_t));
		
                iphdr = (struct ip *) buf;
                iphdrlen = iphdr->ip_hl * 4;    // number of 32-bit words *4 = bytes
		
		/* for some reason, this gets crapped on (prolly has to do with ipv6 wrapping */
		if(from.sin_addr.s_addr == 0)
			from.sin_addr.s_addr =
				iphdr->ip_src.s_addr;
		
                if(bread < iphdrlen + ICMP_MIN)
                        mthis->message("Too few bytes from %s\n",
                                       inet_ntoa(from.sin_addr));

                icmphdr = (struct icmp *) (buf + iphdrlen);

                adr = malloc(25);
		if(!adr)
			err_ret("Could not allocate space");
		
		memset(adr, 0, 25);
		strncpy(adr, inet_ntoa(iphdr->ip_src), 24);
		if((ai = mthis->resolve(adr, NULL, 0, 0)) == NULL)
		{
			len = strlen(adr);
			psuedo = true;
			ai = (struct addrinfo *) malloc(sizeof(struct addrinfo));
			if(!ai)
			{
				err_sys("Memory exhausted\n");
			}
			memset(ai, 0, sizeof(struct addrinfo));
			ai->ai_family = PF_INET;
			ai->ai_protocol = IPPROTO_TCP;
			ai->ai_addrlen = sizeof(struct sockaddr_in);
			ai->ai_addr = (struct sockaddr *) malloc(sizeof(struct sockaddr_in));
			if(!(ai->ai_addr))
				err_sys("memory exhausted\n");
			memset(ai->ai_addr, 0, sizeof(struct sockaddr_in));
			memcpy(ai->ai_addr, &(from.sin_addr), sizeof(struct sockaddr_in));
			ai->ai_canonname = (char *) malloc(len + 1);
			if(!(ai->ai_canonname))
				err_sys("memory exhausted\n");
			memset(ai->ai_canonname, 0, len + 1);
			strncpy(ai->ai_canonname, adr, len);
		}
		free(adr);

                // can't check based on icmp id's, cause only linux keeps these in replies
                // now we check that the sequence number is in range
                if(icmphdr->icmp_seq > seq_no2 + 1 || icmphdr->icmp_seq < low_seq)
                {
                        //keep looking
                        continue;
                }/*
		if(icmphdr->i_id != getpid())
                {
                        //keep looking
                        continue;
                }*/
		/* clock ticks since this packet has been gone */
		triptime = (clock() - icmphdr->icmp_seq); 
		triptime = (triptime * 1000) / CLOCKS_PER_SEC; /* ms */
		/* calculating the trip time this way seems to only have 
		   a resolution of 10 ms */
                if(icmphdr->icmp_type != resptype(my_type))
                {
                        if(icmphdr->icmp_type == ICMP_DEST_UNREACH)
                        {
                                mthis->message("%s: %s(%d) \n",
                                              inet_ntoa(icmphdr->icmp_ip.ip_dst),
                                              unreach_codes(icmphdr->icmp_code),
                                              icmphdr->icmp_code);
                                rec->responses++;
				break;
                        }
                        else if(mthis->sCon.iVerbose > 0)
                        {
                                mthis->message
                                        ("%s[%s]: non-%s recv'd - type: %s code: ",
                                         ai->ai_canonname, inet_ntoa(from.sin_addr),
                                         respstr(resptype(my_type)),
                                         respstr(icmphdr->icmp_type));

                                mthis->output("id: %d -> %s\n", icmphdr->icmp_id,
                                              ai->ai_canonname);

                                continue;
                        }
                }
		
		res.next = NULL;
		len = strlen(ai->ai_canonname);
		
		/* i have problems here, as sometimes this b0rks when its time to get freed */
		res.hostname = (char *) malloc (len +1);
		if(!res.hostname)
			err_sys("Could not malloc!\n");
		memset(res.hostname, 0, len+1);
		strncpy(res.hostname, ai->ai_canonname, len);
		res.ainfo = ai; 
		res.rtt = (double) triptime;
		res.ttl = iphdr->ip_ttl;
		res.user_data = NULL; 
		num_res++;

                mthis->message
                        ("%s[%s] online: icmp_seq = %d. range: %lu ms ttl: %d size: %d\n",
                         ai->ai_canonname, inet_ntoa(from.sin_addr), icmphdr->icmp_seq,
                         triptime, iphdr->ip_ttl, ntohs(iphdr->ip_len));
		/* put the result in the list */
		mthis->insert_results(results, &res);
		/* get rid of our address memory, we dont need it now */
		freeaddrinfo(ai);
		
                if(mthis->pCon.prog != NULL)
		{
			int retv;
			retv = system(mthis->pCon.prog);
			if(WIFSIGNALED(retv) &&
				(WTERMSIG(retv) == SIGINT || WTERMSIG(retv) == SIGQUIT))
			{
				done = 1;
				pthread_exit(NULL);
			}
		}
                rec->responses++;
                continue;
        }
	
	/* these will work even if results are zero */
	mthis->results(results, num_res);
	mthis->free_results(results);
	
        result = rec->hosts - rec->responses;
        if(result > 0)
                mthis->output("no response(s) from %d hosts\n", result);
        if(result < 0)
                mthis->output
                        ("subnet broadcast detected: extra response(s) from %d hosts\n",
                         ~result + 1);
        return NULL;
}


/**
  * @fn checksum(unsigned short *buffer, int size)
  * Compute an ICMP checksum
  * @param buffer Pointer to data
  * @param size Size of data
  * @return Checksum of data
  */
static unsigned short checksum(unsigned short *buffer, int size)
{

        unsigned long cksum = 0;

        while(size > 1)
        {
                cksum += *buffer++;
                size -= sizeof(unsigned short);
        }

        if(size)
        {
                cksum += *(unsigned char *) buffer;
        }

        cksum = (cksum >> 16) + (cksum & 0xffff);
        cksum += (cksum >> 16);
        return (unsigned short) (~cksum);
}

/**
  * @fn icmp_last_error()
  * Return last error code
  * @return Last error code
*/
static int icmp_last_error(void)
{
        return my_errno;
}

/**
  * @fn error(char *x, int y)
  * Display and error and set the error code
  * @param x Message to display
  * @param y Error code to set
*/
static void inline error(char *x, int y)
{
        mthis->message("%s: ERROR %s\n", __FILE__, x);
        my_errno = y;
}

/**
  * @fn resptype(int type)
  * Return the code for the response type 
  * @param type ICMP type sent
  * @return Expected response type
*/
static int resptype(int type)
{
        switch (type)
        {
                case ICMP_ROUTERADVERT:
                        return ICMP_ROUTERSOLICIT;
                case ICMP_ROUTERSOLICIT:
                        return ICMP_ROUTERADVERT;
                case ICMP_TSTAMP:
                        return ICMP_TSTAMPREPLY;
                case ICMP_IREQ:
                        return ICMP_IREQREPLY;
                case ICMP_MASKREQ:
                        return ICMP_MASKREPLY;
                case ICMP_ECHO:
                        return ICMP_ECHOREPLY;
        }
        return -1;
}

/**
  * @fn respstr(int type)
  * Return the name for the response type 
  * @param type ICMP type sent
  * @return Expected response type in a string
*/
static const char *respstr(int type)
{
        switch (type)
        {
                case ICMP_UNREACH:
                        return "dest unreachable";
                case ICMP_SOURCEQUENCH:
                        return "source quench";
                case ICMP_ROUTERADVERT:
                        return "router advert";
                case ICMP_ROUTERSOLICIT:
                        return "router solicit";
                case ICMP_TIMXCEED:
                        return "time exceed";
                case ICMP_PARAMPROB:
                        return "param problem";
                case ICMP_TSTAMP:
                        return "timestamp request";
                case ICMP_TSTAMPREPLY:
                        return "timestamp reply";
                case ICMP_IREQ:
                        return "info request";
                case ICMP_IREQREPLY:
                        return "info reply";
                case ICMP_MASKREQ:
                        return "mask request";
                case ICMP_MASKREPLY:
                        return "mask reply";
                case ICMP_ECHO:
                        return "echo request";
                case ICMP_ECHOREPLY:
                        return "echo reply";
        }
        return "unknown";
}

/**
  * @fn unreach_codes(int code)
  * Return the name of the dest_unreach code 
  * @param code ICMP code number
  * @return ICMP code in a string form
*/
static const char *unreach_codes(int code)
{
        switch (code)
        {
                        // unreach codes:
                case ICMP_UNREACH_NET:
                        return "net unreachable";
                case ICMP_UNREACH_HOST:
                        return "host unreachable";
                case ICMP_UNREACH_PROTOCOL:
                        return "protocol not supported";
                case ICMP_UNREACH_PORT:
                        return "port unreachable";
                case ICMP_UNREACH_NEEDFRAG:
                        return "need fragmentation";
                case ICMP_UNREACH_SRCFAIL:
                        return "source-route failed";
                case ICMP_UNREACH_NET_UNKNOWN:
                        return "network unknown";
                case ICMP_UNREACH_HOST_UNKNOWN:
                        return "host unknown";
                case ICMP_UNREACH_ISOLATED:
                        return "host isolated";
                case ICMP_UNREACH_NET_PROHIB:
                        return "net denied";
                case ICMP_UNREACH_HOST_PROHIB:
                        return "host denied";
                case ICMP_UNREACH_TOSNET:
                        return "bad TOS for net";
                case ICMP_UNREACH_TOSHOST:
                        return "bad TOS for host";
                case ICMP_UNREACH_FILTER_PROHIB:
                        return "filter prohibited";
                case ICMP_UNREACH_HOST_PRECEDENCE:
                        return "host precedence violation";
                case ICMP_UNREACH_PRECEDENCE_CUTOFF:
                        return "precedence cutoff";
        }
        return "unknown";
}

static RETSIGTYPE sig_handler(int i)
{
	switch(i)
	{
	case SIGSEGV:
		err_msg("oops! ACK scan seg faulted.\n");
		abort();
		break;
	}
	/* just close the thread and be done with it */
        pthread_exit(NULL);
}
