/***************************************************************************
    rfc793.c  -  TCP ACK scan plugin
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: rfc793.c,v 1.31 2003/12/06 00:14:04 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file rfc793.c
  * Contains a rfc 793 behavior check (TCP ACK->RST ping)
  */

#include <assert.h>
#include <pthread.h>
#include "rfc793.h"
#include "error.h"
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif

static const char rcsid[] =
        "\100$ autosec/sonar: $Id: rfc793.c,v 1.31 2003/12/06 00:14:04 red0x Exp $";

static int my_socket, err_socket;
static unsigned int num_hosts;
static unsigned char done = 0;
static plugin_t pthis;
static sonar_t *mthis;
static int ready;

static uint16_t my_errno;
static struct timezone tz;
static struct timeval tp;

static int plugin_destroy(void);
static int plugin_args(char opt, char *optarg);
static int plugin_open(void *c_data, int i_data);
static int plugin_close(void);
static void plugin_usage(void);
static void plugin_status(void);
static int plugin_last_error(void);
static int rfc793_sweep(target_t * targets);
static void tvsub(struct timeval *out, struct timeval *in);
static int rfc793_send(target_t * target);
//static int rfc793_recv(host_list_t * hosts);
static void *recv_thread(void *arg);
static void print_packet(struct tcphdr *tcp);


static int in_cksum(unsigned short *addr, int len);
static void add_tcphdr(unsigned char *pkt, struct net_tuple *nt,
                       uint8_t flags);
//dont need this now, but i might later
//static void add_iphdr(unsigned char *pkt, struct net_tuple *nt);
static int send_ipv4(int sox, u_char * pkt, size_t len, struct net_tuple *nt);
static const char *unreach_codes(int code);
static RETSIGTYPE sig_handler(int i);

/** @fn plugin_init(void *in_data)
  * Plugin Init function, part of the Plugin API
  * @param in_data Pointer to a sonar_t structure that will be 
  *  passed from sonar when your plugin is loaded.
  * @return A pointer to a filled plugin_t structure.
  */
void *plugin_init(void *in_data)
{
        assert(in_data != NULL);

        /* useless reference, to avoid a warning */
        if(rcsid)
        {
        }
        /* this is assumed */
        mthis = (sonar_t *) in_data;
        if(mthis->api < 5)
                err_quit("Plugin API must be 5 or higher, update sonar");

        pthis.type = NETWORK_PLUGIN;
        pthis.err = &my_errno;
        pthis.ready = true;
        pthis.version = 5;      /* set API version 5 */
        ready = false;

        pthis.plugin_char = 'A';
        pthis.shortname = "ack scan";
        pthis.longname = "ACK scanning plugin";

        pthis.plugin_init = &plugin_init;
        pthis.plugin_destroy = &plugin_destroy;
        pthis.plugin_args = &plugin_args;
        pthis.plugin_open = &plugin_open;
        pthis.plugin_close = &plugin_close;
        pthis.plugin_next = NULL;
        pthis.plugin_sweep = &rfc793_sweep;
        pthis.plugin_output = NULL;
        pthis.plugin_results = NULL;
        pthis.plugin_usage = &plugin_usage;
        pthis.plugin_status = &plugin_status;
        pthis.plugin_last_error = &plugin_last_error;

        my_errno = PLUGIN_LOADED;

        my_socket = -1;

        num_hosts = 0;

        return &pthis;
}


static int plugin_destroy(void)
{
        mthis = NULL;
        if(my_socket > 0)
        {
                close(my_socket);
                my_socket = -1;
        }

        pthis.type = 0;
        pthis.err = NULL;
        pthis.ready = false;
        ready = false;
        pthis.plugin_init = NULL;
        pthis.plugin_destroy = NULL;
        pthis.plugin_open = NULL;
        pthis.plugin_close = NULL;
        pthis.plugin_next = NULL;
        pthis.plugin_sweep = NULL;
        pthis.plugin_args = NULL;
        pthis.plugin_sweep = NULL;
        pthis.plugin_output = NULL;
        pthis.plugin_results = NULL;
        pthis.plugin_usage = NULL;
        pthis.plugin_status = NULL;
        pthis.plugin_last_error = NULL;
        my_errno = PLUGIN_UNLOADED;
        my_socket = -1;

        return PLUGIN_UNLOADED;
}

/* hopefully, this will eventually leave us */
static int plugin_args(char opt, char *optarg)
{
        return PLUGIN_OK;
}

static int plugin_open(void *c_data, int i_data)
{
        // load socket stuff here
        // i_data will be the TTL to set it to
        uid_t uid;
        struct linger l;

        assert(mthis != NULL);
        assert(mthis->usage != NULL);
        assert(mthis->getprivs != NULL);
        assert(mthis->dropprivs != NULL);

        // setuid root section of code
        {
                uid = mthis->getprivs();
                // root section of code
                //get me  RAW TCP and ICMP sockets
                my_socket = socket(PF_INET, SOCK_RAW, IPPROTO_TCP);
                if(my_socket == -1)
                        err_sys("could not create raw TCP socket");
                err_socket = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
                if(err_socket == -1)
                        err_sys("could not create raw ICMP socket");

                if(setsockopt(my_socket, SOL_IP, IP_TTL, &(mthis->pCon.iTTL),
                              sizeof(mthis->pCon.iTTL)) < 0)
                {
                        *(pthis.err) = PLUGIN_ERROR;
                        err_ret("could not set TTL");
                        //perror("setsockopt");
                        //mthis->usage("could not set TTL");
                }
                l.l_onoff = 0;
                if(mthis->pCon.iTimeout == -1)
                        l.l_linger = 5;
                else
                        l.l_linger = mthis->pCon.iTimeout;
                if(setsockopt(my_socket, SOL_SOCKET, SO_LINGER,
                              (struct linger *) &l,
                              sizeof(struct linger)) < 0)
                {
                        *(pthis.err) = PLUGIN_ERROR;
                        err_ret("could not set timeout");
                        //perror("setsockopt");
                        //mthis->usage("could not set linger time on TCP socket");
                }
                if(setsockopt
                   (err_socket, SOL_SOCKET, SO_LINGER, (struct linger *) &l,
                    sizeof(struct linger)) < 0)
                {
                        *(pthis.err) = PLUGIN_ERROR;
                        err_ret("could not set icmp timeout");
                        //perror("setsockopt");
                        //mthis->usage("could not set linger time on ICMP socket");
                }
                if(mthis->pCon.bSockDebug)
                {
                        l.l_onoff = 1;
                        if(setsockopt(my_socket, SOL_SOCKET, SO_DEBUG,
                                      (int *) &(l.l_onoff), sizeof(int)) < 0)
                        {
                                *(pthis.err) = PLUGIN_ERROR;
                                err_ret("could not set debug mode");
                                //perror("setsockopt");
                                //exit(-1);
                        }
                }
                if(mthis->pCon.bBroadcast)
                {
                        l.l_onoff = 1;
                        if(setsockopt(my_socket, SOL_SOCKET, SO_BROADCAST,
                                      (int *) &(l.l_onoff), sizeof(int)) < 0)
                        {
                                *(pthis.err) = PLUGIN_ERROR;
                                err_ret("could not set broadcast mode");
                                //perror("setsockopt");
                                //exit(-1);
                        }
                }

                mthis->dropprivs(uid);
        }
        //back to normalcy

        ready = true;
        num_hosts = mthis->num_targets;
        return PLUGIN_OK;
}

static int plugin_close(void)
{
        close(my_socket);
        pthis.ready = false;

        return PLUGIN_OK;
}

static void plugin_usage(void)
{
        assert(mthis != NULL);
        assert(mthis->output != NULL);
        mthis->output("\n ACK Scan plugin: requires root priveleges\n");
        mthis->output("  Usage: hostname:port/mask ...\n");
        mthis->output
                ("  If no port is specified, this will scan *all* ports\n");
        mthis->output("  in /etc/services\n");
        mthis->output("    Example: sonar -sA www.google.com:80/32\n");
        exit(-1);
}

static void plugin_status(void)
{
        assert(mthis != NULL);
        assert(mthis->output != NULL);
        if(pthis.ready && ready)
        {
                mthis->output("  ACK Scan plugin ready::\n");
                mthis->output("  TTL %d :: delay %d :: ", mthis->pCon.iTTL,
                              mthis->pCon.iPostDelay);
        }
}

static int plugin_last_error(void)
{
        return my_errno;
}

static int rfc793_sweep(target_t * targets)
{
        unsigned int counter = 0, i;
        pthread_t tid;
        pthread_attr_t attr;
        struct recv_t recd;
        int timeout;

        assert(targets != NULL);

        // just make sure we, and sonar, are ready for processing
        if(!pthis.ready || !ready)
        {
                err_msg("plugin not ready\n");
                return PLUGIN_ERROR;
        }

        timeout = mthis->pCon.iTimeout;
        if(timeout <= 0)
                timeout = 5000;

        pthread_attr_init(&attr);
        memset(&recd, 0, sizeof(struct recv_t));

        recd.socket = my_socket;
        recd.errs = err_socket;
        /* if we are broadcasting, increase the responses to collect */
        /* and increase the timeout value */
        recd.hosts = num_hosts;
        if(mthis->pCon.bBroadcast > 0 && targets->extra_hosts == 0)
                recd.hosts += (unsigned int) mthis->pCon.bBroadcast * 255;

        recd.start = clock();
        /* set the default timeout values */
        recd.max =
                recd.start + (timeout * recd.hosts * CLOCKS_PER_SEC) / 1000;
        /* add time for a per host delay */
        if(mthis->pCon.iHostDelay > 0)
                recd.max +=
                        (mthis->pCon.iHostDelay * recd.hosts *
                         CLOCKS_PER_SEC) / 1000;
        /* add a default magic number to get some extra (late) responses */
        recd.max += CLOCKS_PER_SEC / 2000;      /* default is one half extra second */

        /* start the receiving thread */
        done = 0;
        if(pthread_create(&tid, &attr, recv_thread, &recd) != 0)
                err_sys("could not create thread");

        while(targets != NULL && counter < num_hosts)
        {
                if(targets->num_ports > 0)
                {
                        recd.max +=
                                (timeout * (targets->num_ports - 1) *
                                 CLOCKS_PER_SEC) / 1000;
                        recd.hosts += (targets->num_ports - 1);
                        targets->num_ports = 0;
                }
                for(i = 0; i < 65536; i++)
                {
                        /* stupid hax0r tr1ckz */
                        if(targets->ports[i] == PORT_SCAN)
                        {
                                targets->ports[i] = PORT_SCANNING;
                                if(targets->extra_hosts > 0)
                                {
                                        recd.max +=
                                                (timeout *
                                                 targets->extra_hosts *
                                                 CLOCKS_PER_SEC) / 1000;
                                        recd.hosts += targets->extra_hosts;
                                        if(mthis->sCon.iVerbose > 0)
                                                err_msg("Increasing timeout, netmask detected: %d", recd.max);
                                }
                                if(rfc793_send(targets) == PLUGIN_ERROR)
                                {
                                        return PLUGIN_ERROR;
                                }
                                counter++;
                                /* sleep after this host (allows more dest unreachables to come back) */
                                if(counter < num_hosts)
                                        usleep(mthis->pCon.iHostDelay * 1000);
                        }
                }
                targets = targets->next;
        }
        done = 1;
        /* wait for the thread to exit */
        pthread_join(tid, NULL);

        if(mthis->pCon.iPostDelay)
        {
                /* sleep this many milliseconds after this scan */
                usleep(mthis->pCon.iPostDelay * 1000);
        }
        return PLUGIN_OK;
}

/* @fn rfc793_send(target_t *target)
 * Send a specially crafted TCP ACK packet to first entry in *target
 * @param target Filled target_t pointer
 * @return PLUGIN_OK if all went well
 * @todo Timestamp each packet individually
 */
static int rfc793_send(target_t * target)
{
        unsigned char *packet = NULL;
        struct addrinfo *ai = NULL;
        struct net_tuple nt;
        struct sockaddr_in *sa = NULL;
        struct in_addr ia;
        char name[256];
        int bwrote, i;
        unsigned int j;
        int total_size = sizeof(struct tcphdr);
	int my_ports[65547];

        assert(target != NULL);
        assert(target->entry != NULL);
        assert(target->addrinfo != NULL);
        assert(target->addrinfo->ai_addr != NULL);
        assert(my_socket != 0);
        assert(my_socket != -1);
        assert(mthis->output != NULL);

        srand(getpid() ^ (uid_t) time(NULL));
        gethostname(name, 255);
        ai = mthis->resolve(name, NULL, 0, PF_INET);
	memcpy(my_ports, target->ports, 65537 * sizeof(int));
        for(j = 0; j <= target->extra_hosts; j++)
        {
                packet = (unsigned char *) malloc(total_size);
                if(!packet)
                {
                        err_ret("could not allocate space for packet\n");
                        return PLUGIN_ERROR;
                }
                memset(packet, 0, total_size);
                memset(name, 0, 256);


                if(ai)
                {
                        memcpy(&nt.src, ai->ai_addr, ai->ai_addrlen);
                }
                else
                        return PLUGIN_RESOLVE;


                if(mthis->pCon.iSrcPort != -1)
                        nt.sport = htons(mthis->pCon.iSrcPort);
                else
                        nt.sport =
                                htons(1 +
                                      (int) (65535.0 * rand() /
                                             (RAND_MAX + 1.0)));

                switch (target->addrinfo->ai_family)
                {
                        case PF_INET:
                                sa = (struct sockaddr_in *) target->
                                        addrinfo->ai_addr;
                                ia = sa->sin_addr;
                                ia.s_addr = htonl(ntohl(ia.s_addr) + j);
                                nt.dst = ia.s_addr;
                                break;
                        default:
                                err_quit("address family not supported yet\n");
                }

                for(i = 0; i < 65536; i++)
                {
                        /* lame hax0r tr1ck (1 = not scanned, 2 = scanning, 3 = scanned) */

                        if(target->ports[i] == PORT_SCANNING)
                        {
                                target->ports[i] = PORT_SCANNED;
                                nt.dport = htons(i);
                        }
                }


                add_tcphdr(packet, &nt, TH_ACK);
                // packet is ready to send       

                //time stamp our packet, for round trip time
                gettimeofday(&(tp), &(tz));

                bwrote = send_ipv4(my_socket, packet, total_size, &nt);
                if(bwrote == -1)
                {
                        err_ret("send_ipv4 failed");
                        //perror("send_ipv4");
                        return PLUGIN_ERROR;
                }
                if(bwrote < (int) total_size)
                {
                        err_msg("wrote %d bytes\n", bwrote);
                }
                free(packet);
		/* fresh copy of port list */
		memcpy(target->ports, my_ports, 65537);
        }

        return PLUGIN_OK;
}

static void *recv_thread(void *arg)
{
        fd_set rfds;
        struct ip *iphdr;
        struct icmp *icmphdr;
        struct tcphdr *tcp = NULL;
        unsigned short iphdrlen;
        unsigned long triptime, unreach = 0, filt = 0, unfilt = 0;
        struct timeval tv;
        int bread, ret;
        char buf[MAX_PACKET], *adr = NULL;
        struct recv_t *rec = (struct recv_t *) arg;
        int fromlen, my_socket = rec->socket, err_socket =
                rec->errs, result, retv;
        struct sockaddr_in from;
        static struct addrinfo *ai = NULL;
        int psuedo = false;
        result_t *results = NULL,       /* results used to pass back to sonar */
                *first_res = NULL;      /* first on the list */
        static result_t res;    /* temporary for *this* result */
        unsigned long num_res = 0;      /* number of results */

        assert(my_socket != -1 && my_socket != 0);
        assert(err_socket != -1 && err_socket != 0);
        assert(mthis->resolve != NULL);

	/* signals, baby */
	signal(SIGTERM, sig_handler);
        signal(SIGINT, sig_handler);
        signal(SIGHUP, sig_handler);
	signal(SIGSEGV, sig_handler);
        /* make us a list */
        results = (result_t *) malloc(sizeof(result_t));
        if(!results)
                err_sys("No memory for result list.\n");
        first_res = results;    /* store a pointer to the top of the list */
        memset(results, 0, sizeof(result_t));   /* dont want to accidentally free an uninitialized value */

        rec->responses = 0;
        while((done == 0 || clock() <= rec->max)
              && rec->responses < rec->hosts)
        {
                /* dont do anything up here, or we'll eat CPU */
		tv.tv_usec = 0;
                tv.tv_sec = 0;
                FD_ZERO(&rfds);
                FD_SET(my_socket, &rfds);
                FD_SET(err_socket, &rfds);
                ret = select(((my_socket >
                               err_socket) ? my_socket + 1 : err_socket + 1),
                             &rfds, NULL, NULL, &tv);
                if(ret == -1)
                {
                        err_ret("could not select socket\n");
                        return NULL;    //PLUGIN_ERROR;
                }
                if(ret == 0)
                {
                        continue;       //return PLUGIN_NORESP;
                }
                if(psuedo == true)
                {
                        assert(ai != NULL);
                        psuedo = false;
                        free(ai);
                }
                /* XXX: breaks result returning */
                /*else
                   freeaddrinfo(ai); */

                memset(&from, 0, sizeof(struct sockaddr_in));

                tv.tv_sec = 0;
                tv.tv_usec = 0;

                memset(buf, 0, MAX_PACKET);
                /* clear the result */
                memset(&res, 0, sizeof(result_t));
                if(FD_ISSET(my_socket, &rfds))
                {
                        /* read our packet */
                        bread = recvfrom(my_socket, buf, MAX_PACKET, 0,
                                         (struct sockaddr *) &(from),
                                         (socklen_t *) & (fromlen));
                        if(bread == -1)
                        {
                                //perror("recvfrom");
                                err_ret("recvfrom error");
                                continue;       //return PLUGIN_ERROR;
                        }
                        /* XXX: timestamping should be on a per-packet basis */
                        gettimeofday(&tv, &(tz));

                        tvsub(&tv, &tp);
                        triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);

                        iphdr = (struct ip *) buf;
                        iphdrlen = iphdr->ip_hl * 4;    // number of 32-bit words *4 = bytes

                        if(bread < iphdrlen + sizeof(struct tcphdr))
                                mthis->message("Too few bytes from %s\n",
                                               inet_ntoa(from.sin_addr));

                        tcp = (struct tcphdr *) (buf + iphdrlen);

                        if(from.sin_addr.s_addr == 0)
                                from.sin_addr.s_addr = iphdr->ip_src.s_addr;

                        adr = malloc(25);
                        if(!adr)
                                err_ret("Could not allocate space");

                        memset(adr, 0, 25);
                        strncpy(adr, inet_ntoa(iphdr->ip_src), 24);
                        if((ai = mthis->resolve(adr, NULL, 0, 0)) == NULL)
                        {
                                int len = strlen(adr);
                                psuedo = true;
                                ai = (struct addrinfo *)
                                        malloc(sizeof(struct addrinfo));
                                if(!ai)
                                {
                                        err_sys("Memory exhausted\n");
                                }
                                memset(ai, 0, sizeof(struct addrinfo));
                                ai->ai_family = PF_INET;
                                ai->ai_protocol = IPPROTO_TCP;
                                ai->ai_addrlen = sizeof(struct sockaddr_in);
                                ai->ai_addr =
                                        (struct sockaddr *)
                                        malloc(sizeof(struct sockaddr_in));
                                if(!(ai->ai_addr))
                                        err_sys("memory exhausted\n");
                                memset(ai->ai_addr, 0,
                                       sizeof(struct sockaddr_in));
                                memcpy(ai->ai_addr, &(from.sin_addr),
                                       sizeof(struct sockaddr_in));
                                ai->ai_canonname = (char *) malloc(len + 1);
                                if(!(ai->ai_canonname))
                                        err_sys("Memory exhausted\n");
                                memset(ai->ai_canonname, 0, len + 1);
                                strncpy(ai->ai_canonname, adr, len);
                        }
                        free(adr);
                        // we get something other than a RST, it should be filtered
                        if((tcp->th_flags & TH_RST))
                        {
                                int len;
                                res.next = NULL;
                                len = strlen(ai->ai_canonname);

                                /*res.hostname = ai->ai_canonname; */
                                res.hostname = (char *) malloc(len + 1);
                                if(!res.hostname)
                                        err_sys("Could not malloc!\n");
                                memset(res.hostname, 0, len + 1);
                                strncpy(res.hostname, ai->ai_canonname, len);
                                res.ports[ntohs(tcp->th_sport)] =
                                        PORT_UNFILTERED;
                                res.ainfo = ai;
                                res.rtt = (double) triptime;
                                res.ttl = iphdr->ip_ttl;
                                res.user_data = NULL;
                                num_res++;
                                
                                mthis->message
                                        ("%s[%s]:%d online & UNfiltered: range: %lu ms ttl: %d size: %d\n",
                                         res.hostname,
                                         inet_ntoa(from.sin_addr),
                                         ntohs(tcp->th_sport), triptime,
                                         iphdr->ip_ttl, ntohs(iphdr->ip_len));
                                /* put the result in the list */
                                mthis->insert_results(results, &res);
                                /* get rid of our address memory, we dont need it now */
                                freeaddrinfo(ai);
                                /* print the packet if they want it */
                                if(mthis->sCon.iVerbose > 1)
                                        print_packet(tcp);
                                if(mthis->pCon.prog != NULL)
                                {
                                        retv = system(mthis->pCon.prog);
                                        if(WIFSIGNALED(retv) &&
                                           (WTERMSIG(retv) == SIGINT
                                            || WTERMSIG(retv) == SIGQUIT))
                                        {
                                                done = 1;
                                                pthread_exit(NULL);
                                        }
                                }

                                rec->responses++;
                                unfilt++;
                                continue;
                        }


                        continue;
                }
                else if(FD_ISSET(err_socket, &rfds))
                {
                        int len;
                        bread = recvfrom(err_socket, buf, MAX_PACKET, 0,
                                         (struct sockaddr *) &(from),
                                         (socklen_t *) & (fromlen));
                        if(bread == -1)
                        {
                                //perror("recvfrom");
                                err_ret("recvfrom error");
                                return NULL;    //PLUGIN_ERROR;
                        }

                        gettimeofday(&tv, &(tz));

                        tvsub(&tv, &tp);
                        triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);

                        iphdr = (struct ip *) buf;
                        iphdrlen = iphdr->ip_hl * 4;    // number of 32-bit words *4 = bytes

                        if(bread < iphdrlen + sizeof(struct icmp))
                                mthis->message("Too few bytes from %s\n",
                                               inet_ntoa(from.sin_addr));

                        icmphdr = (struct icmp *) (buf + iphdrlen);

                        tcp = (struct tcphdr *) (icmphdr +
                                                 sizeof(struct icmp));

                        if(icmphdr->icmp_type != ICMP_DEST_UNREACH)
                        {
                                if(mthis->sCon.iVerbose > 1)
                                        mthis->message
                                                ("bad ICMP response from host, abnormal behavior\n");
                                continue;       //return PLUGIN_OK;
                                // Parse non-unreachable ICMP returns
                        }
                        /* change the address information to reflect where this packet was headed */
                        from.sin_addr = icmphdr->icmp_ip.ip_dst;

                        adr = malloc(25);
                        if(!adr)
                                err_ret("Could not allocate space");

                        memset(adr, 0, 25);
                        strncpy(adr, inet_ntoa(from.sin_addr), 24);
                        if((ai = mthis->resolve(adr, NULL, 0, 0)) == NULL)
                        {
                                len = strlen(adr);
                                psuedo = true;
                                ai = (struct addrinfo *)
                                        malloc(sizeof(struct addrinfo));
                                if(!ai)
                                {
                                        err_sys("Memory exhausted\n");
                                }
                                memset(ai, 0, sizeof(struct addrinfo));
                                ai->ai_family = PF_INET;
                                ai->ai_protocol = IPPROTO_TCP;
                                ai->ai_addrlen = sizeof(struct sockaddr_in);
                                ai->ai_addr =
                                        (struct sockaddr *)
                                        malloc(sizeof(struct sockaddr_in));
                                memcpy(ai->ai_addr, &(from.sin_addr),
                                       sizeof(struct sockaddr_in));
                                ai->ai_canonname = (char *) malloc(len + 1);
                                memset(ai->ai_canonname, 0, len + 1);
                                strncpy(ai->ai_canonname, adr, len);
                        }
                        free(adr);
                        rec->responses++;
                        switch (icmphdr->icmp_code)
                        {
                                case ICMP_UNREACH_PORT:
                                        assert(ai != NULL);
                                        res.next = NULL;
                                        /*res.hostname = ai->ai_canonname; */
                                        len = strlen(ai->ai_canonname);
                                        res.hostname =
                                                (char *) malloc(len + 1);
                                        if(!res.hostname)
                                                err_sys("Could not malloc!\n");
                                        memset(res.hostname, 0, len + 1);
                                        strncpy(res.hostname,
                                                ai->ai_canonname, len);
                                        res.ports[ntohs(tcp->th_dport)] =
                                                PORT_FILTERED;
                                        res.ainfo = ai;
                                        res.rtt = (double) triptime;
                                        res.ttl = iphdr->ip_ttl;
                                        res.user_data = NULL;
                                        num_res++;
                                        mthis->insert_results(results, &res);
                                        mthis->message
                                                ("%s[%s]:%d online & filtered: range: %lu ms ttl: %d size: %d\n",
                                                 ai->ai_canonname,
                                                 inet_ntoa(from.sin_addr),
                                                 ntohs(tcp->th_dport),
                                                 triptime, iphdr->ip_ttl,
                                                 ntohs(iphdr->ip_len));
                                        filt++;
                                        break;
                                case ICMP_UNREACH_HOST:
                                        assert(ai != NULL);
                                        if(mthis->sCon.iVerbose > 0)
                                                mthis->message
                                                        ("%s[%s] host unreachable\n",
                                                         ai->ai_canonname,
                                                         inet_ntoa(from.
                                                                   sin_addr),
                                                         triptime,
                                                         iphdr->ip_ttl,
                                                         ntohs(iphdr->
                                                               ip_len));
                                        unreach++;
                                        break;
                                default:
                                        assert(ai != NULL);
                                        if(mthis->sCon.iVerbose > 1)
                                        {
                                                mthis->message
                                                        ("%s[%s]: destination unreachable: %s\n",
                                                         ai->ai_canonname,
                                                         inet_ntoa(from.
                                                                   sin_addr),
                                                         unreach_codes
                                                         (icmphdr->
                                                          icmp_code));
                                        }
                                        unreach++;
                                        break;
                        }
                        /* get rid of our address memory, we dont need it now */
                        freeaddrinfo(ai);
                        continue;
                }
        }

        if(mthis->sCon.iVerbose > 2)
        {
                result = rec->hosts - rec->responses;
                if(result > 0)
                        mthis->output("no response(s) from %d hosts\n",
                                      result);
                if(result < 0)
                        mthis->output
                                ("subnet broadcast detected: extra response(s) from %d hosts\n",
                                 ~result + 1);
                mthis->message("ACK Scan complete:\n");
                mthis->message
                        ("  Scanned %d hosts/ports, %d hosts online, %d filtered, %d unfiltered, %d hosts unreachable\n",
                         rec->hosts, rec->responses - unreach, filt, unfilt,
                         unreach);
        }
        mthis->results(results, num_res);
        mthis->free_results(results);
        return NULL;
}

/**
  * @fn tvsub(register struct timeval *out, register struct timeval *in)
  * Subtract one struct timeval from another
  * @param out Pointer to destination (will be filled with out - in)
  * @param in Pointer to src
*/
static void tvsub(struct timeval *out, struct timeval *in)
{
        if((out->tv_usec -= in->tv_usec) < 0)
        {
                out->tv_sec--;
                out->tv_usec += 1000000;
        }
        out->tv_sec -= in->tv_sec;
}

static void print_packet(struct tcphdr *tcp)
{
        mthis->output("  sport: %d dport: %d\n", ntohs(tcp->th_sport),
                      ntohs(tcp->th_dport));
        mthis->output("  flags: (");
        if(tcp->th_flags & TH_FIN)
                mthis->output("F");
        else
                mthis->output("-");
        if(tcp->th_flags & TH_SYN)
                mthis->output("S");
        else
                mthis->output("-");
        if(tcp->th_flags & TH_RST)
                mthis->output("R");
        else
                mthis->output("-");
        if(tcp->th_flags & TH_PUSH)
                mthis->output("P");
        else
                mthis->output("-");
        if(tcp->th_flags & TH_ACK)
                mthis->output("A");
        else
                mthis->output("-");
        if(tcp->th_flags & TH_URG)
                mthis->output("U");
        else
                mthis->output("-");
        mthis->output(") offset: %d window: %d seq: %d ack: %d\n",
                      tcp->th_off, tcp->th_win, tcp->th_seq, tcp->th_ack);
}

int in_cksum(unsigned short *addr, int len)
{
        int nleft = len;
        int sum = 0;
        u_short *w = addr;
        u_short answer = 0;


        while(nleft > 1)
        {
                sum += *w++;
                nleft -= 2;
        }


        if(nleft == 1)          /* padding */
        {
                *(u_char *) (&answer) = *(u_char *) w;
                sum += answer;
        }


        return (sum);
}


void add_tcphdr(unsigned char *pkt, struct net_tuple *nt, uint8_t flags)
{
        struct tcphdr tcp;
        struct _fakehead fakehead;
        int sum;


        memset(&tcp, 0, sizeof(tcp));
        memset(&fakehead, 0, sizeof(fakehead));


        tcp.th_dport = nt->dport;
        tcp.th_sport = nt->sport;
        fakehead.saddr = nt->src;
        fakehead.daddr = nt->dst;
        fakehead.zero = 0, fakehead.protocol = 6;
        fakehead.tot_len = htons(TCP_SIZE);
        sum = in_cksum((u_short *) & fakehead, sizeof(fakehead));
        tcp.th_off = TCP_SIZE >> 2;
        tcp.th_seq = 1 + (int) (65535.0 * rand() / (RAND_MAX + 1.0));
        tcp.th_ack = 1 + (int) (65535.0 * rand() / (RAND_MAX + 1.0));
        tcp.th_flags |= flags;  /* ADD the flags */
        tcp.th_win = 0;
        sum += in_cksum((u_short *) & tcp, sizeof(tcp));
        tcp.th_sum = CKSUM_CARRY(sum);
        memcpy(pkt, &tcp, sizeof(tcp));
}


/* 
 * send out ipv4-packet 
 * with data 'pkt' of length 'len' 
 * returns the number of characters sent, or -1 if an error occured 
 */
int send_ipv4(int sox, u_char * pkt, size_t len, struct net_tuple *nt)
{
        struct sockaddr_in to;


        to.sin_family = AF_INET;
        memcpy(&to.sin_addr.s_addr, &nt->dst, sizeof(u_long));


        return (sendto
                (sox, pkt, len, 0, (struct sockaddr *) &to, sizeof(to)));
}

/**
  * @fn unreach_codes(int code)
  * Return the name of the dest_unreach code 
  * @param code ICMP code number
  * @return ICMP code in a string form
*/
static const char *unreach_codes(int code)
{
        switch (code)
        {
                        // unreach codes:
                case ICMP_UNREACH_NET:
                        return "net unreachable";
                case ICMP_UNREACH_HOST:
                        return "host unreachable";
                case ICMP_UNREACH_PROTOCOL:
                        return "protocol not supported";
                case ICMP_UNREACH_PORT:
                        return "port unreachable";
                case ICMP_UNREACH_NEEDFRAG:
                        return "need fragmentation";
                case ICMP_UNREACH_SRCFAIL:
                        return "source-route failed";
                case ICMP_UNREACH_NET_UNKNOWN:
                        return "network unknown";
                case ICMP_UNREACH_HOST_UNKNOWN:
                        return "host unknown";
                case ICMP_UNREACH_ISOLATED:
                        return "host isolated";
                case ICMP_UNREACH_NET_PROHIB:
                        return "net denied";
                case ICMP_UNREACH_HOST_PROHIB:
                        return "host denied";
                case ICMP_UNREACH_TOSNET:
                        return "bad TOS for net";
                case ICMP_UNREACH_TOSHOST:
                        return "bad TOS for host";
                case ICMP_UNREACH_FILTER_PROHIB:
                        return "filter prohibited";
                case ICMP_UNREACH_HOST_PRECEDENCE:
                        return "host precedence violation";
                case ICMP_UNREACH_PRECEDENCE_CUTOFF:
                        return "precedence cutoff";
        }
        return "unknown";
}

static RETSIGTYPE sig_handler(int i)
{
	switch(i)
	{
	case SIGSEGV:
		err_msg("oops! ACK scan seg faulted.\n");
		abort();
		break;
	}
	/* just close the thread and be done with it */
        pthread_exit(NULL);
}
