/***************************************************************************
                          mutex.c  -  contains debuggable mutexes
                             -------------------
    begin                : Tues May 13 2003
    copyright            : (C) 2003 by red0x
    email                : red0x@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file mutex.c
  * File contains debuggable mutexs
  */

#include "mutex.h"
#include "error.h"
#include "log.h"
#include <string.h>

static const char rcsid[] =
        "\100$ autosec/sonar: $Id: mutex.c,v 1.3 2003/11/04 23:48:40 red0x Exp $";
/* @fn sonar_mutex_init(sonar_mutex_t * mutex, const pthread_mutexattr_t * mutexattr, char *id)
 * Initialize a pthread mutex in a wrapped way that stores an identifier for debugging
 * @param mutex A sonar_mutex_t pointer
 * @param mutexattr A pthread_mutexattr_t pointer
 * @param id User readable mutex ID string
 * @return the same value of pthread_mutex_init
 */
int sonar_mutex_init(sonar_mutex_t * mutex,
                     const pthread_mutexattr_t * mutexattr, char *id)
{
#ifdef DEBUG
	int len = strlen(id);
        strncpy(mutex->id, id, (len > 80 ? 80 : len));
#endif

        return pthread_mutex_init(&mutex->mutex, mutexattr);
}

/* @fn sonar_mutext_lock(sonar mutext_t *mutex, char *who)
 * Try to lock the mutex, print a message if it blocks.
 * If DEBUG is not defined, it just wraps pthread_mutext_lock directly
 * @param mutex sonar_mutex_t to try to lock
 * @param who name of what function is trying to lock the mutex
 * @return same values as pthread_mutex_lock
*/
int sonar_mutex_lock(sonar_mutex_t * mutex, char *who)
{

#ifndef DEBUG

        return pthread_mutex_lock(&mutex->mutex);

#else

        if(pthread_mutex_trylock(&mutex->mutex))
        {
                err_msg("sonar_mutex: BLOCK when %s tried to lock mutex %s because it is locked by %s. continue trying...)\n", who, mutex->id, mutex->locked_by);

                pthread_mutex_lock(&mutex->mutex);
        }

        log_write(LOG_DBUG, "  sonar_mutex: %s has now locked mutex %s\n", who, mutex->id);
        mutex->locked_by = who;

        return 1;

#endif
}

/* @fn sonar_mutex_unlock(sonar_mutex_t * mutex, char *who)
 * Try to unlock the mutex.
 * @param mutex Mutex to unlock
 * @param who Name of function trying to unlock
 * @return same values as pthread_mutext_unlock
 */ 
int sonar_mutex_unlock(sonar_mutex_t * mutex, char *who)
{

        log_write(LOG_DBUG, "  sonar_mutex: mutex %s unlocked by %s\n", mutex->id, who);
        return pthread_mutex_unlock(&mutex->mutex);
}

/* @fn sonar_mutex_destroy(sonar_mutex_t * mutex)
 * Directly calls pthread_mutex_destroy
 * @param mutex Mutex to destroy
 * @return same as pthread_mutex_destroy
 */
int sonar_mutex_destroy(sonar_mutex_t * mutex)
{
        return pthread_mutex_destroy(&mutex->mutex);
}
