/***************************************************************************
   plugin.c  -  plugin API
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: plugin.c,v 1.24 2003/11/07 08:44:35 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file plugin.c
  * File contains plugin API Implementation\n
  * Plugin API version 1.0 \n
  * To make a plugin, look in plugin.h for the plugin_t structure,
  * and make sure your plugins can provide all of the needed methods.
  * If you dont need a method, for now, just make it return PLUGIN_OK
  * Until I get a more detailed API documentation, that is how we will
  * do it.
  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "plugin.h"
#include "util.h"
#include "error.h"
#include "log.h"
#include <assert.h>

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif


#include <ltdl.h>


#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

static const char rcsid[] =
        "\100$ autosec/sonar: $Id: plugin.c,v 1.24 2003/11/07 08:44:35 red0x Exp $";

/*woutput(char *msg, ...);
extern int verbose;*/
static int plugin_add(int type, plugin_t * p_data, sonar_t * mthis, void *hptr);

/// list of input plugins
plugin_list_t input_list[MAX_INP];     
/// list of output plugins
plugin_list_t output_list[MAX_OUT];     
/// list of network plugins
plugin_list_t network_list[MAX_NET];    

/// plugin id system PID
static int first_pid;
/// Number of input plugins
unsigned long int num_inp,
        /// number of output plugins
  num_out,
        /// number of network plugins
  num_net;

/** @name Plugin API Internals
  * These are behind the scenes functions that your plugins, 
  * nor you, should ever need to call.
  * Do not call these functions!
 */
//@{

/** @fn plugin_ready()
  * Plugin API Internal -- DO NOT CALL FROM PLUGINS\n
  * Checks if the plugin subsystem is ready.  Mainly useful if plugin_lists are dynamically allocated
  * As they are not, this function is pretty much pointless.
  * @return 1 if the system is ready
  */
static int inline plugin_ready()
{
        // are they all defined?
        if(!input_list || !output_list || !network_list)
                return 0;
        return 1;
}

/** @fn plugin_init(plugin_list_t * x)
  * Plugin API Internal -- DO NOT CALL FROM PLUGINS\n
  * Sets the plugin_list_t member "plug" to NULL
  * This function only exists in case we need to do more
  * stuff here later.
  * @param plugin_list plugin_list_t pointer
  */
static void inline plugin_init(plugin_list_t * plugin_list)
{
        // set the plugin to NULL
        plugin_list->plug = NULL;
	plugin_list->handle = NULL;
}

/** @fn plugin_newid()
  * Plugin API Internal -- DO NOT CALL FROM PLUGINS\n
  * Gets a new pluginID for your plugin, once loaded
  * @return A new, unique plugin ID
  */
static int inline plugin_newid()
{
        return (unsigned long int) ++first_pid; /* (first_pid += ((unsigned int)
                                                  (200.0 * rand() /
                                                   (RAND_MAX + 64.0))));*/
}

//@}

/** @name Plugin API -- Sonar side
  * These are behind the scenes functions that sonar uses
  * to load and initialize your plugins.  You should never
  * need access to these functions.
  * Do not call these functions!
 */
//@{

/**
  * @fn plugin_start(sonar_t *mthis)
  * Startup and load the plugin API
  * @param mthis Pointer to a filled sonar_t structure with all of sonar's
  * pertinent function pointers and information
  * @return PLUGIN_OK if ok, PLUGIN_ERROR otherwise
  */
int plugin_start(sonar_t * mthis)
{
        if(!input_list)
        {
                err_sys("Input Array not Initialized: %s line %d", __FILE__,
                        __LINE__);
                return PLUGIN_ERROR;
        }
        plugin_init(&input_list[0]);
        if(!output_list)
        {
                err_sys("Output Array not Initialized: %s line %d",
                        __FILE__, __LINE__);
                return PLUGIN_ERROR;
        }
        plugin_init(&output_list[0]);
        if(!network_list)
        {
                err_sys("Network Array not Initialized: %s line %d",
                        __FILE__, __LINE__);
                return PLUGIN_ERROR;
        }
        plugin_init(&network_list[0]);
        first_pid = 0;
        num_inp = num_out = num_net = 0;
        return PLUGIN_OK;
}

/**
  * @fn plugin_end()
  * Unload the plugin API
  * @return PLUGIN_OK
  */
int plugin_end()
{
#ifdef	DEBUG
        err_msg("plugin_end();");
#endif
        first_pid = 0;
        num_inp = num_out = num_net = 0;
        return PLUGIN_OK;
}

/**
  * @fn plugin_loadall(sonar_t *mthis)
  * Find and load all plugins in the plugin directory
  * @param mthis Pointer to a filled sonar_t structure
  * @return PLUGIN_OK if ok, PLUGIN_ERROR otherwise
  * @todo Lock the plugin library while checking permissions, to avoid a race
  */
int plugin_loadall(sonar_t * mthis)
{
        /* directory to search defined as PLUGINDIR */
        DIR *dir;
        lt_dlhandle plugin;
        void *(*plugin_init) (void *);
	struct stat *my_buf;
	/* try to get the owner and mode of sonar */
	if((my_buf = get_stats(mthis->my_name)) == NULL)
		err_sys("could not stat %s", mthis->my_name);

        log_write(LOG_DBUG, "  opendir(%s);\n", PLUGINDIR);
        dir = opendir(PLUGINDIR);
        if(dir)
        {
                struct dirent *entry;
		log_write(LOG_DBUG, "  readdir(%p);\n", dir);

                while((entry = readdir(dir)) != NULL)
                {

                        char str[1024];
			struct stat stat_buf;
                        int len = strlen(entry->d_name);
			log_write(LOG_DBUG, "  found %s\n", entry->d_name);
			/* all plugins must have this in their names, at the beginning */
			if(len < strlen("libsonar"))
				continue;
			/* this code filters out other libraries
			 needed once we went libtool */
			if(strncmp(entry->d_name, "libsonar", strlen("libsonar")) != 0)
				continue;  /* not the one */
                        if((entry->d_name[len - 3] == '.') &&
                           (entry->d_name[len - 2] == 's') &&
                           (entry->d_name[len - 1] == 'o'))
                        {
                                /* found it */
                                memset(str, 0, 1024);
                                snprintf(str, 1023, "%s/%s", PLUGINDIR,
                                         entry->d_name);

				/* check permissions */
				/* XXX: lock, no interrupts to avoid race */
				log_write(LOG_DBUG, "  stat(%s, %p);\n",str, &stat_buf);
				if(stat(str, &stat_buf) < 0)
					err_sys("could not stat %s", str);
				if(stat_buf.st_mode &= S_IWOTH)
				{
					log_write(LOG_CONSOLE | LOG_SLOG, "CRITICAL: %s is world writtable! not loading!", str);
					continue;
				}
				if(stat_buf.st_uid != my_buf->st_uid)
				{
					/* if they are owned by root, its ok */
					if(stat_buf.st_uid != 0)
					{
						log_write(LOG_CONSOLE | LOG_SLOG, "CRITICAL: %s owned by someone else (%d, not %d)! not loading!", str, stat_buf.st_uid, my_buf->st_uid);
						continue;
					}
				}
				/* XXX: unlock */
				/* end check */
				lt_dlinit();
				log_write(LOG_DBUG, "  lt_dlopen(%s); => ", str);
                                if((plugin = lt_dlopen(str)) == NULL)
                                {
                                        char *dl_error_msg =
                                                (char *) lt_dlerror();
					log_write(LOG_DBUG, "  FAILED!\n");
                                        log_write(LOG_CONSOLE | LOG_SLOG, "could not load %s: %s\n",
                                                str, dl_error_msg);
                                        continue;
                                }
				log_write(LOG_DBUG, "   %p (ok)\n", plugin);
				log_write(LOG_DBUG, "  lt_dlsym(%p, \"plugin_init\"); => ",
                                        plugin);
                                plugin_init =
                                        (void *(*)(void *)) lt_dlsym(plugin,
                                                                  "plugin_init");
                                if(plugin_init != NULL)
                                {
                                        plugin_t *plug;
					log_write(LOG_DBUG, "  %p (ok)\n", plugin_init);
                                        log_write(LOG_DBUG, "  plugin_init(%p) => ",
                                                &mthis);
                                        plug = (plugin_t *)
                                                plugin_init(mthis);
                                        if(!plug)
                                        {
						log_write(LOG_DBUG, "  FAILED!\n");
						log_write(LOG_CONSOLE | LOG_SLOG, "  Could not load %s:\nplugin_init failed\n", str);
                                        }
                                        else
                                        {
						log_write(LOG_DBUG, "  %p (ok)\n", plug);
						/* only support from version 4 on up */
						if(plug->version > API_VER || plug->version < 4)
						{
							log_write(LOG_CONSOLE | LOG_SLOG, "  Warning: Plugin \"%s\" wanted an unsupported api, version \'%d\', not loading\n", plug->shortname, plug->version);
							lt_dlclose(plugin);
							continue;
						}
						log_write(LOG_DBUG, "  plugin_add(%d, %p, %p, %p);\n", plug, plug->type, plug, mthis, plugin);
                                                plugin_add(plug->type, plug,
                                                           mthis, plugin);
                                        }
                                }
                                else
                                {
                                        char *dl_error_msg =
                                                (char *) lt_dlerror();
					log_write(LOG_DBUG, "  FAILED!\n");
                                        log_write(LOG_CONSOLE | LOG_SLOG, "  could not load plugin_init in %s:\n%s\n", str, dl_error_msg);
                                }
                        }
                }
        }
	
#ifdef CLOSEDIR_VOID
        closedir(dir);
#else
	if(closedir(dir) < 0)
		err_ret("closedir failure");
#endif

        return PLUGIN_OK;
}

/**
  * @fn plugin_killall(void)
  * Unload all plugins
  * @return PLUGIN_OK if ok, PLUGIN_ERROR otherwise
  */
int plugin_killall(void)
{
	int err = PLUGIN_OK;
	log_write(LOG_DBUG, "  plugin_killall();\n");
        
        //get all input, output, network plugins
        //call the plugin_destroy function
        for(; num_inp > 0; num_inp--)
        {
                int my_err;
                assert(input_list[num_inp - 1].plug != NULL);
                assert(input_list[num_inp - 1].plug->plugin_destroy != NULL);
                //if plugin is also an output plugin, let that call its destroy function
                if(!(input_list[num_inp - 1].plug->type & OUTPUT_PLUGIN))
                {
			log_write(LOG_DBUG, "  unloading %s\n",
                                input_list[num_inp - 1].plug->shortname);
                        log_write(LOG_DBUG, "  input_list[%d].plug->plugin_destroy();\n",
                                num_inp - 1);

                        if((my_err =
                            input_list[num_inp - 1].plug->plugin_destroy()) !=
                           PLUGIN_UNLOADED)
                        {
				log_write(LOG_DBUG, "  plugin not unloaded\n");
                                err = my_err;
                        }
                        input_list[num_inp - 1].plug = NULL;
                        lt_dlclose(input_list[num_inp - 1].handle);
                }
        }
        for(; num_out > 0; num_out--)
        {
                int my_err;
                assert(output_list[num_out - 1].plug != NULL);
                assert(output_list[num_out - 1].plug->plugin_destroy != NULL);
		assert(output_list[num_out - 1].plug->plugin_close != NULL);

                //if plugin is also a network plugin, let that call it destroy function
                if(!(output_list[num_out - 1].plug->type & NETWORK_PLUGIN))
                {
			log_write(LOG_DBUG, "  unloading %s\n",
                                output_list[num_out - 1].plug->shortname);
                        log_write(LOG_DBUG, "  output_list[%d].plug->plugin_destroy();\n",
                                num_out - 1);
			if(output_list[num_out -1].plug->ready)
				output_list[num_out -
                                        1].plug->plugin_close();
                        if((my_err =
                            output_list[num_out -
                                        1].plug->plugin_destroy()) !=
                           PLUGIN_UNLOADED)
                        {
				log_write(LOG_DBUG, "  plugin not unloaded");
                                err = my_err;
                        }
                        output_list[num_out - 1].plug = NULL;
                        lt_dlclose(output_list[num_out - 1].handle);
                }
        }
        for(; num_net > 0; num_net--)
        {
                int my_err;
                assert(network_list[num_net - 1].plug != NULL);
                assert(network_list[num_net - 1].plug->plugin_destroy !=
                       NULL);
		log_write(LOG_DBUG, "  unloading %s\n",
                        network_list[num_net - 1].plug->shortname);
                log_write(LOG_DBUG, "  network_list[%d].plug->plugin_destroy();\n",
                        num_net - 1);
                if((my_err =
                    network_list[num_net - 1].plug->plugin_destroy()) !=
                   PLUGIN_UNLOADED)
                {
			log_write(LOG_DBUG, "  plugin not unloaded\n");
                        err = my_err;
                }
                network_list[num_net - 1].plug = NULL;
                lt_dlclose(network_list[num_net - 1].handle);
        }
	lt_dlexit();
        return err;
}

/**
  * @fn plugin_add(int type, plugin_t *p_data, sonar_t *mthis, void *hptr)
  * Add plugin to proper plugin_list
  * @param type Plugin type (ex, for input and output, it would be INPUT_PLUGIN | OUTPUT_PLUGIN
  * @param p_data Pointer to filled plugin_t structure
  * @param mthis Pointer to a filled sonar_t structure
  * @param hptr Handle returned by lt_dlopen(3)
  * @return PLUGIN_OK if ok, PLUGIN_ERROR otherwise
  */
static int plugin_add(int type, plugin_t * p_data, sonar_t * mthis, void *hptr)
{
        if(!plugin_ready())
        {
                log_write(LOG_CONSOLE | LOG_SLOG, "plugin_add: plugins not ready");
                return PLUGIN_ERROR;
        }
        
	assert(p_data != NULL);
	assert(mthis != NULL);
	assert(hptr != NULL);
        
        if(type & INPUT_PLUGIN)
        {
                plugin_init(&(input_list[num_inp]));

                input_list[num_inp].plug = p_data;
                input_list[num_inp].handle = hptr;
                input_list[num_inp].plug->pluginID = plugin_newid();
		log_write(LOG_DBUG, "  input_list[%d].plug->pluginID = %d\n", num_inp,
                        input_list[num_inp].plug->pluginID);
                num_inp++;
                if(num_inp > MAX_INP)
                {
                        err_quit("TOO MANY INPUT PLUGINS, ABORTING!");
                }
        }
        if(type & OUTPUT_PLUGIN)
        {
                plugin_init(&(output_list[num_out]));

                output_list[num_out].plug = p_data;
                output_list[num_out].handle = hptr;
                output_list[num_out].plug->pluginID = plugin_newid();
		log_write(LOG_DBUG, "  output_list[%d].plug->pluginID = %d\n", num_out,
                        output_list[num_out].plug->pluginID);
                num_out++;
                if(num_out > MAX_OUT)
                {
                        err_quit("TOO MANY OUTPUT PLUGINS, ABORTING...");
                }
        }
        if(type & NETWORK_PLUGIN)
        {
                plugin_init(&(network_list[num_net]));

                network_list[num_net].plug = p_data;
                network_list[num_net].handle = hptr;
                network_list[num_net].plug->pluginID = plugin_newid();
		log_write(LOG_DBUG, "  network_list[%d].plug->pluginID = %d\n", num_net,
                        network_list[num_net].plug->pluginID);
                num_net++;
                if(num_net > MAX_NET)
                {
                        err_quit("TOO MANY NETWORK PLUGINS, ABORTING...");
                }
        }
        return PLUGIN_OK;
}

//@}

/** @name Plugin API -- Hooks
  * These are hook functions used to print usage and status information
  * as well as handling command line options
  * Do not call these functiions!
 */
//@{

/**
  * @fn plugin_status_hook(void)
  * Call each plugin's status function to print extra status reports
  * at the status screen
  */
void plugin_status_hook(void)
{
        unsigned int plug;
        //get all input, output, network plugins
        //call the plugin_status function

        for(plug = 0; plug < num_inp; plug++)
        {
                {
                        assert(input_list[plug].plug != NULL);
                        assert(input_list[plug].plug->plugin_status != NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                input_list[plug].plug->plugin_status();
        }
        for(plug = 0; plug < num_out; plug++)
        {
                {
                        assert(output_list[plug].plug != NULL);
                        assert(output_list[plug].plug->plugin_status != NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                output_list[plug].plug->plugin_status();
        }
        for(plug = 0; plug < num_net; plug++)
        {
                {
                        assert(network_list[plug].plug != NULL);
                        assert(network_list[plug].plug->plugin_status !=
                               NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                network_list[plug].plug->plugin_status();
        }
        return;
}


/**
  * @fn plugin_command_handler(char opt, char *optarg)
  * Call each plugin's plugin_args function to see if it handles this option
  * @param opt Option char as passed from getopt
  * @param optarg Option argument as passed from getopt
  * @return SONAR_ERROR if no plugins handled the argument\n
  * SONAR_OK if someone handled it
  */
int plugin_command_handler(char opt, char *optarg)
{
        //go through the plugins, and see if this option is handled, first network
        // plugins, then input, then output
        // else, return SONAR_ERROR
        unsigned int plug = 0;
        int error = PLUGIN_ERROR;
        for(plug = 0; plug < num_inp; plug++)
        {
                int myerror;
                assert(input_list[plug].plug != NULL);
                assert(input_list[plug].plug->plugin_args != NULL);
                if((myerror =
                    input_list[plug].plug->plugin_args(opt,
                                                       optarg)) !=
                   PLUGIN_ERROR && error == PLUGIN_ERROR)
                        error = myerror;
        }
        for(plug = 0; plug < num_out; plug++)
        {
                int myerror;
                assert(output_list[plug].plug != NULL);
                assert(output_list[plug].plug->plugin_args != NULL);
                if((myerror =
                    output_list[plug].plug->plugin_args(opt,
                                                        optarg)) !=
                   PLUGIN_ERROR && error == PLUGIN_ERROR)
                        error = myerror;
        }
        for(plug = 0; plug < num_net; plug++)
        {
                int myerror;
                assert(network_list[plug].plug != NULL);
                assert(network_list[plug].plug->plugin_args != NULL);
                if((myerror =
                    network_list[plug].plug->plugin_args(opt,
                                                         optarg)) !=
                   PLUGIN_ERROR && error == PLUGIN_ERROR)
                        error = myerror;
        }
        return error;
}

/**
  * @fn plugin_message_hook(char *msg, va_list *ap)
  * Call each output plugin's output function if its ready
  * @param msg Message format
  * @param ap va_list of arguments
  */
void plugin_message_hook(char *msg, va_list * ap)
{
        unsigned int plug;
        int err_val = 0;
        //get all output plugins
        //call the plugin_output function

        for(plug = 0; plug < num_out; plug++)
        {
                {
                        assert(output_list[plug].plug != NULL);
                        assert(output_list[plug].plug->type & OUTPUT_PLUGIN);
                        assert(output_list[plug].plug->plugin_output != NULL);
                }
                if(output_list[plug].plug->ready == true)
                {
                        err_val =
                                output_list[plug].plug->plugin_output(msg,
                                                                      ap);
                        if(err_val < 0)
                                err_msg("error: %d, output plugin error",
                                        err_val);
                }
        }
        return;
}

//@}


void clear_sonar_context(struct sonarContext *in)
{
        assert(in != NULL);
        memset((void *) in, 0, sizeof(struct sonarContext));
        in->oFile = stdout;
        in->iCount = DEFAULT_COUNT;
	in->bRandom = 0;
}

void clear_plugin_context(struct pluginContext *in)
{
        assert(in != NULL);
        memset((void *) in, 0, sizeof(struct pluginContext));
        in->iTimeout = DEFAULT_TIMEOUT;
	in->iPostDelay = 0;
        in->iTTL = 255;
        in->iSrcPort = -1;
	in->prog = NULL;
	in->iHostDelay = 0;
}
