/***************************************************************************
                          target.c  -  contains functions to add targets
                             -------------------
    begin                : Tues May 13 2003
    copyright            : (C) 2003 by red0x
    email                : red0x@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file target.c
  * File contains functions to add targets
  */

#include <assert.h>
#include <netdb.h>
#include "target.h"
#include "service.h"
#include "net.h"
#include "error.h"
#include "log.h"
#include "sonar.h"

extern sonar_t mthis;

static const char rcsid[] =
        "\100$ autosec/sonar: $Id: target.c,v 1.8 2003/12/05 22:47:52 red0x Exp $";

static unsigned long check_netmask(target_t *target);
static int parse_target_ports(char *ports, target_t **target);
static int load_default_ports(target_t **target);

int sonar_add_target(sonar_t * this, char *entry)
{
        struct addrinfo *addr;
        target_t *target, *ent;
	char *hostname;
	char *ports;
	char *mask;
        int len,i=0;
	unsigned int num_ports = 0, num_hosts = 0;
	int imask = 32;

        assert(this != NULL);
        assert(entry != NULL);

	len = strlen(entry);
	
	// copy the entry over
	hostname = (char *) malloc(len + 1);
	if(!hostname)
		err_sys("sonar_add_target: malloc error");
	memset((char *) hostname, 0, len + 1);
	strncpy(hostname, entry, len);
	ports = mask = hostname;

	// split up the hostname and ports
	while(ports[0] != ':' && i < len)
	{
		*ports++;
		i++;
	}
	i = 0;
	while(mask[0] != '/' && i < len)
	{
		*mask++;
		i++;
	}
	// null delimit the hostname;
	if(*ports == ':')
		*ports++ = 0x00;
	else
		ports = NULL;
	//otherwise, leave it alone, assume default ports
	if(*mask == '/')
		*mask++ = 0x00;
	else
		mask = NULL;
	// ports now contains ports
	// mask now contains a hostmask

	log_write(LOG_DBUG, "  hostname: %s, ports: %s mask: %s\n", hostname, ports, mask);

	if(mask)
	{
		imask = atoi(mask);
		if(imask < 0 || imask > 32)
		{
			err_msg("Bad hostmask: %d, using /32\n", imask);
			imask = 32;
		}
	}
        target = (target_t *) malloc(sizeof(target_t));
        if(target == NULL)
                err_sys("sonar_add_target: malloc error");
	// start filling in needed values
	target->entry = (char *) malloc(len + 1);
	if(!target->entry)
		err_sys("sonar_add_target: malloc error");
	memset(target->entry, 0,len + 1);
	strncpy(target->entry, entry, len);
	target->netmask = imask;
	//parse list of ports.
	num_ports = parse_target_ports(ports, &(target));

	// See if we have an ipv4 hostname, and wrap it

        // call Wrapped host_serv so we dont have to check for errors.
        addr = Host_serv(hostname, NULL, 0, 0);
        assert(addr != NULL);
        target->addrinfo = addr;
	target->num_ports = num_ports;
        target->next = NULL;
	num_hosts = check_netmask(target);

	ent = this->vectors;
	if(!ent)
		this->vectors = target;
	else
	{
		while(ent->next != NULL)
			ent= ent->next;
		ent->next = target;
	}

	this->num_targets++;
	/* XXX: Fix this, put it in target_t */
	this->num_ports += num_ports;

        return SONAR_OK;
}

static unsigned long check_netmask(target_t *target)
{
	unsigned long retval = 0, longtmp;
	uint32_t start, end;
	struct sockaddr_in *sa;

	assert(target != NULL);
	assert(target->addrinfo != NULL);
	assert(target->addrinfo->ai_addr != NULL);
	sa = (struct sockaddr_in *) target->addrinfo->ai_addr;
	longtmp = ntohl(sa->sin_addr.s_addr);
	
	start = longtmp & (unsigned long) (0 - (1 << (32 - target->netmask)));
	if(((start << 24) >> 24) == 0)
		mthis.pCon.bBroadcast++;
	end = longtmp | (unsigned long) ((1<<(32 - target->netmask)) - 1);
	if(((end << 24) >> 24) == 0xff)
		mthis.pCon.bBroadcast++;
	if(start > end)
		err_msg("Invalid netmask\n");
	retval = end - start + 1;
	if(retval <= 0)
		err_quit("Invalid hostmask\n");
	log_write(LOG_DBUG, "  Netmask: %d, hosts %u\n", target->netmask, retval);
	target->extra_hosts = retval - 1;
	/* update the target to be the start of the range */
	sa->sin_addr.s_addr = htonl(start);
	return retval;
}

static int parse_target_ports(char *ports, target_t **target)
{
	int i=0, len;
	char *dest, *next;
	unsigned int ret = 0;
	
	if(!*target){
		*target = malloc(sizeof(target_t));
		if(!*target)
			err_sys("parse_target_ports: malloc error");
	}
	memset((*target)->ports, PORT_DONTSCAN, 65537);
	/* get default ports from /etc/services */
	if(!ports)
	{
		return load_default_ports(target);
	}
	
	len = strlen(ports);
	
	// all ports
	if(strchr(ports, '*') != NULL)
	{
		log_write(LOG_DBUG, "  setting all ports to scan\n");
		for( i = 0; i < 65536; i++)
		{
			(*target)->ports[i] = PORT_SCAN;
			ret++;
		}
		return ret;
	}

	log_write(LOG_DBUG, "  clearing port list\n");
	//clear the port list
	for(i = 0; i < 65537; i++)
		(*target)->ports[i] = 0;

	dest = strtok(ports, ",");

	while(dest)
	{
		
		// check for range
		next = strchr(dest, '-');
		if(!next) // no range
		{
			sonar_resolve_service(dest, *target);
			ret++;
		}
		else //has a range
		{
			// dest has beginning
			// get the last range
			*next++ = '\0';
			ret = sonar_resolve_service_r(dest, next, *target);
			
		}
		
		
		dest = strtok(NULL, ",");
			
	}
	return ret; 
}

static int load_default_ports(target_t **target)
{
	struct servent *se;
	assert(*target != NULL);
	while((se = getservent()) != NULL)
	{
		(*target)->ports[se->s_port] = 1;
	}
	endservent();
	return SONAR_OK;
	
}

#ifdef DEBUG
int sonar_debug_targets(target_t *target)
{
	int i = 0;
	while(target)
	{
		err_msg(" target debug:");
		err_msg("  entry: %s", target->entry);
		err_msg("  hostname: %s", target->addrinfo->ai_canonname);
		err_msg("  ports:");
		for(i = 0; i < 65536; i++)
			if(target->ports[i] == 1)
				printf("%d,",i);
		err_msg("");
		target = target->next;
	}

	return SONAR_OK;
}
#endif //DEBUG
